from ecoscope.mapping import (
    EcoMap,
    FloatElement,
    NorthArrowElement,
    ScaleElement,
    GeoTIFFElement,
    PrintControl,
)
import ee
import geopandas
import pytest
from branca.element import MacroElement
from branca.colormap import StepColormap
from folium.raster_layers import TileLayer, ImageOverlay
from folium.map import FitBounds
from bs4 import BeautifulSoup


def test_ecomap_base():
    m = EcoMap()
    assert len(m._children) == 7


def test_repr_html():
    m = EcoMap(width=800, height=600)

    soup = BeautifulSoup(m._repr_html_(), "html.parser")
    assert soup.iframe.get("width") == "800"
    assert soup.iframe.get("height") == "600"

    soup = BeautifulSoup(m._repr_html_(fill_parent=True), "html.parser")
    assert soup.iframe.get("width") == "100%"
    assert soup.iframe.get("height") == "100%"

    assert m._parent.width == 800 and m._parent.height == 600


def test_add_legend():
    m = EcoMap()
    # Test that we can assign hex colors with or without #
    m.add_legend(legend_dict={"Black_NoHash": "000000", "White_Hash": "#FFFFFF"})
    assert len(m.get_root()._children) == 2
    assert isinstance(list(m.get_root()._children.values())[1], MacroElement)
    html = m._repr_html_()
    assert "Black_NoHash" in html
    assert "White_Hash" in html


def test_add_north_arrow():
    m = EcoMap()
    m.add_north_arrow()
    assert len(m._children) == 8
    assert isinstance(list(m._children.values())[7], NorthArrowElement)
    assert "&lt;svg" in m._repr_html_()


def test_add_scale_bar():
    m = EcoMap()
    m.add_scale_bar()
    assert len(m._children) == 8
    assert isinstance(list(m._children.values())[7], ScaleElement)
    assert "&lt;svg" in m._repr_html_()


def test_add_title():
    m = EcoMap()
    m.add_title("THIS IS A TEST TITLE")
    assert len(m._children) == 8
    assert isinstance(list(m._children.values())[7], FloatElement)
    assert "THIS IS A TEST TITLE" in m._repr_html_()


@pytest.mark.skipif(not pytest.earthengine, reason="No connection to EarthEngine.")
def test_add_ee_layer_image():
    m = EcoMap()
    vis_params = {"min": 0, "max": 4000, "opacity": 0.5, "palette": ["006633", "E5FFCC", "662A00", "D8D8D8", "F5F5F5"]}
    ee_object = ee.Image("USGS/SRTMGL1_003")
    m.add_ee_layer(ee_object, vis_params, "DEM")
    assert len(m._children) == 8
    assert isinstance(list(m._children.values())[7], TileLayer)
    assert "earthengine.googleapis.com" in m._repr_html_()


@pytest.mark.skipif(not pytest.earthengine, reason="No connection to EarthEngine.")
def test_add_ee_layer_image_collection():
    m = EcoMap()
    vis_params = {"min": 0, "max": 4000, "opacity": 0.5}
    ee_object = ee.ImageCollection("MODIS/006/MCD43C3")
    m.add_ee_layer(ee_object, vis_params, "DEM")
    assert len(m._children) == 8
    assert isinstance(list(m._children.values())[7], TileLayer)
    assert "earthengine.googleapis.com" in m._repr_html_()


@pytest.mark.skipif(not pytest.earthengine, reason="No connection to EarthEngine.")
def test_add_ee_layer_feature_collection():
    m = EcoMap()
    vis_params = {"min": 0, "max": 4000, "opacity": 0.5, "palette": ["006633", "E5FFCC", "662A00", "D8D8D8", "F5F5F5"]}
    ee_object = ee.FeatureCollection("LARSE/GEDI/GEDI02_A_002/GEDI02_A_2021244154857_O15413_04_T05622_02_003_02_V002")
    m.add_ee_layer(ee_object, vis_params, "DEM")
    assert len(m._children) == 8
    assert isinstance(list(m._children.values())[7], TileLayer)
    assert "earthengine.googleapis.com" in m._repr_html_()


def test_zoom_to_bounds():
    m = EcoMap()
    m.zoom_to_bounds((34.683838, -3.173425, 38.869629, 0.109863))
    assert len(m._children) == 8
    assert isinstance(list(m._children.values())[7], FitBounds)
    assert "[[-3.173425, 34.683838], [0.109863, 38.869629]]" in m._repr_html_()


def test_zoom_to_gdf():
    m = EcoMap()
    x1 = 34.683838
    y1 = -3.173425
    x2 = 38.869629
    y2 = 0.109863
    gs = geopandas.GeoSeries.from_wkt(
        [f"POINT ({x1} {y1})", f"POINT ({x2} {y1})", f"POINT ({x1} {y2})", f"POINT ({x2} {y2})"]
    )
    gs = gs.set_crs("EPSG:4326")
    m.zoom_to_gdf(gs)
    assert len(m._children) == 8
    assert isinstance(list(m._children.values())[7], FitBounds)
    assert "[[-3.173425, 34.683838], [0.109863, 38.869629]]" in m._repr_html_()


def test_add_local_geotiff():
    m = EcoMap()
    m.add_local_geotiff("tests/sample_data/raster/uint8.tif", cmap=None)
    assert len(m._children) == 8
    assert isinstance(list(m._children.values())[7], GeoTIFFElement)
    assert "new GeoRasterLayer" in m._repr_html_()


def test_add_local_geotiff_with_cmap():
    m = EcoMap()
    m.add_local_geotiff("tests/sample_data/raster/uint8.tif", cmap="jet")
    assert len(m._children) == 9
    assert isinstance(list(m._children.values())[7], StepColormap)
    assert isinstance(list(m._children.values())[8], GeoTIFFElement)
    html = m._repr_html_()
    assert "new GeoRasterLayer" in html
    assert "d3.scale" in html


def test_add_print_control():
    m = EcoMap()
    m.add_print_control()
    assert len(m._children) == 8
    assert isinstance(list(m._children.values())[7], PrintControl)
    assert "L.control.BigImage()" in m._repr_html_()


@pytest.mark.parametrize(
    "file, geom_type",
    [
        ("tests/sample_data/vector/maec_4zones_UTM36S.gpkg", "tests/sample_data/vector/observations.geojson"),
        ("polygon", "point"),
    ],
)
def add_datashader_gdf(file, geom_type):
    m = EcoMap()
    gdf = geopandas.GeoDataFrame.from_file(file)
    m.add_datashader_gdf(gdf, geom_type, zoom=False)
    assert len(m._children) == 8
    assert isinstance(list(m._children.values())[7], ImageOverlay)
    assert "L.imageOverlay(" in m._repr_html_()


def add_datashader_gdf_with_zoom():
    m = EcoMap()
    gdf = geopandas.GeoDataFrame.from_file("tests/sample_data/vector/maec_4zones_UTM36S.gpkg")
    m.add_datashader_gdf(gdf, "polygon", zoom=True)
    assert len(m._children) == 9
    assert isinstance(list(m._children.values())[7], ImageOverlay)
    assert isinstance(list(m._children.values())[8], FitBounds)
    assert "L.imageOverlay(" in m._repr_html_()



import os
import re

from setuptools import setup

version_file = open("ecoscope/version.py").read()
version_data = dict(re.findall(r'__([a-z]+)__\s*=\s*"([^"]+)"', version_file))

try:
    descr = open(os.path.join(os.path.dirname(__file__), "README.rst")).read()
except IOError:
    descr = ""

dependencies = [
    "affine",
    "astroplan",
    "backoff",
    "datashader",
    "earthengine-api",
    "earthranger-client",
    "folium",
    "geopandas",
    "igraph",
    "kaleido",
    "mapclassify",
    "matplotlib",
    "networkx",
    "numba",
    "plotly",
    "pyarrow",
    "pyproj",
    "pytest",
    "pytest-cov",
    "pytest-mock",
    "pytest-reportlog",
    "rasterio",
    "scipy",
    "scikit-image",
    "scikit-learn",
    "selenium",
    "tqdm",
    "xyzservices",
]

setup(
    name="ecoscope",
    version=version_data["version"],
    description="Standard Analytical Reporting Framework for Conservation",
    long_description=descr,
    url="http://github.com/wildlife-dynamics/ecoscope",
    author="Jake Wall",
    author_email="walljcg@gmail.com",
    license="MIT",
    packages=[
        "ecoscope",
        "ecoscope.analysis",
        "ecoscope.analysis.UD",
        "ecoscope.base",
        "ecoscope.contrib",
        "ecoscope.io",
        "ecoscope.mapping",
        "ecoscope.plotting",
    ],
    platforms="Posix; MacOS X; Windows",
    install_requires=dependencies,
    zip_safe=False,
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    include_package_data=True,
)



import geopandas as gpd
import numpy as np
import pandas as pd
from shapely.geometry import box


def create_meshgrid(
    aoi,
    in_crs,
    out_crs,
    xlen=1000,
    ylen=1000,
    return_intersecting_only=True,
    align_to_existing=None,
):
    """Create a grid covering `aoi`.

    Parameters
    ----------
    aoi : shapely.geometry.base.BaseGeometry
        The area of interest. Should be in a UTM CRS.
    in_crs : value
        Coordinate Reference System of input `aoi`. Can be anything accepted by `pyproj.CRS.from_user_input()`.
        Geometry is automatically converted to UTM CRS as an intermediate for computation.
    out_crs : value
        Coordinate Reference System of output `gs`. Can be anything accepted by `pyproj.CRS.from_user_input()`.
        Geometry is automatically converted to UTM CRS as an intermediate for computation.
    xlen : int, optional
        The width of a grid cell in meters.
    ylen : int, optional
        The height of a grid cell in meters.
    return_intersecting_only : bool, optional
        Whether to return only grid cells intersecting with the aoi.
    align_to_existing : geopandas.GeoSeries or geopandas.GeoDataFrame, optional
        If provided, attempts to align created grid to start of existing grid. Requires a CRS and valid geometry.

    Returns
    -------
    gs : geopandas.GeoSeries
        Grid of boxes. CRS is converted to `out_crs`.
    """

    a = gpd.array.from_shapely([aoi], crs=in_crs)
    int_crs = a.estimate_utm_crs()
    aoi = a.to_crs(int_crs)[0]
    del a

    bounds = aoi.bounds

    x1 = bounds[0] - bounds[0] % xlen
    y1 = bounds[1] - bounds[1] % ylen
    x2 = bounds[2]
    y2 = bounds[3]

    if align_to_existing is not None:
        align_to_existing = align_to_existing.geometry  # Treat GeoDataFrame as GeoSeries
        assert not align_to_existing.isna().any()
        assert not align_to_existing.is_empty.any()

        align_to_existing = align_to_existing.to_crs(int_crs)

        total_existing_bounds = align_to_existing.total_bounds

        x1 += total_existing_bounds[0] % xlen - xlen
        y1 += total_existing_bounds[1] % ylen - ylen

    boxes = [box(x, y, x + xlen, y + ylen) for x in np.arange(x1, x2, xlen) for y in np.arange(y1, y2, ylen)]

    gs = gpd.GeoSeries(boxes, crs=int_crs)
    if return_intersecting_only:
        gs = gs[gs.intersects(aoi)]

    return gs.to_crs(out_crs)


class cachedproperty:  # noqa
    """
    The ``cachedproperty`` is used similar to :class:`property`, except
    that the wrapped method is only called once. This is commonly used
    to implement lazy attributes.
    """

    def __init__(self, func):
        self.func = func

    def __doc__(self):
        return getattr(self.func, "__doc__")

    def __isabstractmethod__(self):
        return getattr(self.func, "__isabstractmethod__", False)

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value

    def __repr__(self):
        cls = self.__class__.__name__
        return f"<{cls} func={self.func}>"


def groupby_intervals(df, col, intervals):
    """
    Parameters
    ----------
    df : pd.DataFrame
        Data to group
    col : str
        Name of column to group on
    intervals : pd.IntervalIndex
        Intervals to group on

    Returns
    -------
    pd.core.groupby.DataFrameGroupBy

    """

    if not df[col].is_monotonic_increasing:
        df = df.sort_values(col)

    return pd.concat(
        {
            interval: df.iloc[left_i:right_i]
            for interval, left_i, right_i in zip(
                intervals,
                df[col].searchsorted(intervals.left, side="left" if intervals.closed in ("left", "both") else "right"),
                df[col].searchsorted(
                    intervals.right, side="right" if intervals.closed in ("right", "both") else "left"
                ),
            )
        }
    ).groupby(level=0)


def create_interval_index(start, intervals, freq, overlap=pd.Timedelta(0), closed="right", round_down_to_freq=False):
    """
    Parameters
    ----------
    start : str or datetime-like
        Left bound for creating IntervalIndex
    intervals : int, optional
        Number of intervals to create
    freq : str, Timedelta or DateOffset
        Length of each interval
    overlap : Timedelta or DateOffset, optional
        Length of overlap between intervals
    closed : {"left", "right", "both", "neither"}, default "right"
        Whether the intervals are closed on the left-side, right-side, both or neither
    round_down_to_freq : bool
        Start will be rounded down to freq

    Returns
    -------
    interval_index : pd.IntervalIndex

    """

    freq = pd.tseries.frequencies.to_offset(freq)
    overlap = pd.tseries.frequencies.to_offset(overlap)
    if round_down_to_freq:
        start = start.round(freq)

    left = [start]
    for i in range(1, intervals):
        start = start + freq
        left.append(start - i * overlap)
    left = pd.DatetimeIndex(left)

    return pd.IntervalIndex.from_arrays(left=left, right=left + freq, closed=closed)


class ModisBegin(pd._libs.tslibs.offsets.SingleConstructorOffset):
    """
    Primitive DateOffset to support MODIS period start times.
    """

    def apply(self, other: pd.Timestamp) -> pd.Timestamp:
        assert other.tz is not None, "Timestamp must be timezone-aware"
        other = other.astimezone("UTC").round("d")
        for i in range(self.n):
            other = min(other + pd.offsets.YearBegin(), other + pd.DateOffset(days=(16 - (other.dayofyear - 1) % 16)))
        return other


def create_modis_interval_index(start, intervals, overlap=pd.Timedelta(0), closed="right"):
    """
    Parameters
    ----------
    start : str or datetime-like
        Left bound for creating IntervalIndex. Will be rounded up to next start of MODIS period
    intervals : int, optional
        Number of intervals to create
    overlap : Timedelta or DateOffset, optional
        Length of overlap between intervals
    closed : {"left", "right", "both", "neither"}, default "right"
        Whether the intervals are closed on the left-side, right-side, both or neither

    Returns
    -------
    interval_index : pd.IntervalIndex

    """

    start = start + ModisBegin()

    left = [start]
    for i in range(1, intervals):
        start = start + ModisBegin()
        left.append(start - i * overlap)
    left = pd.DatetimeIndex(left)

    return pd.IntervalIndex.from_arrays(left=left, right=left + pd.Timedelta(days=16), closed=closed)


def add_val_index(df, index_name, val):
    """
    Parameters
    ----------
    df : pd.DataFrame
        Dataframe to create index on
    index_name : str
        Name of index
    val : str
        Current column to rename and set as index

    Returns
    -------
    pd.DataFrame

    """
    if val in df.columns:
        return df.rename(columns={val: index_name}).set_index(index_name, append=True)
    else:
        df[index_name] = val
        df = df.set_index(index_name, append=True)
    return df


def add_temporal_index(df, index_name, time_col, directive):
    """
    Parameters
    ----------
    df : pd.DataFrame
        Dataframe to create index on
    index_name : str
        Name of temporal index
    time_col : str
        Name of the time column
    directive : str
        Time format string

    Returns
    -------
    pd.DataFrame

    """
    if directive and time_col:
        df[index_name] = df[time_col].dt.strftime(directive)
        return df.set_index(index_name, append=True)
    else:
        return df



import geopandas as gpd
import pandas as pd
import pytest
from shapely.geometry import Point, box

from ecoscope.base import EcoDataFrame


class TestEcoDataFrame(object):
    def setup_method(self):
        self.gdf = gpd.GeoDataFrame(
            {
                "id": [1, 2],
                "my_geometry": [box(1, 1, 2, 2), box(3, 3, 4, 4)],
                "other_geometry": [Point(0, 0), Point(1, 1)],
            },
            geometry="my_geometry",
            crs="epsg:4326",
        )

    def test_constructor(self):
        edf = EcoDataFrame(self.gdf)
        assert isinstance(edf, EcoDataFrame)
        assert edf._constructor == EcoDataFrame

    def test_crs_gdf(self):
        edf = EcoDataFrame(self.gdf)
        assert edf.crs == "epsg:4326"

    def test_crs_custom(self):
        edf = self.gdf.copy()
        edf.crs = None
        edf = EcoDataFrame(edf, crs="epsg:3857")
        assert edf.crs == "epsg:3857"

    def test_to_crs(self):
        edf = EcoDataFrame(self.gdf)
        edf = edf.to_crs("epsg:3857")
        assert isinstance(edf, EcoDataFrame)
        assert edf.crs == "epsg:3857"

    def test_geometry_gdf(self):
        edf = EcoDataFrame(self.gdf)
        assert edf.geometry.name == "my_geometry"

    def test_geometry_custom(self):
        edf = EcoDataFrame(self.gdf, geometry="other_geometry")
        assert edf.geometry.name == "other_geometry"

    def test_set_geometry(self):
        edf = EcoDataFrame(self.gdf)
        edf = edf.set_geometry("other_geometry")
        assert isinstance(edf, EcoDataFrame)
        assert edf.geometry.name == "other_geometry"

    def test_from_file(self, mocker):
        mock = mocker.patch.object(gpd.GeoDataFrame, "from_file")
        edf = EcoDataFrame.from_file("file_name.geojson")
        mock.assert_called_once_with("file_name.geojson")
        assert isinstance(edf, EcoDataFrame)

    def test_from_features(self, mocker):
        mock = mocker.patch.object(gpd.GeoDataFrame, "from_features")
        edf = EcoDataFrame.from_features("__features__")
        mock.assert_called_once_with("__features__")
        assert isinstance(edf, EcoDataFrame)

    def test_getitem_series(self):
        edf = EcoDataFrame(self.gdf)
        edf = edf[edf.id > 1]
        assert isinstance(edf, EcoDataFrame)
        assert len(edf) == 1

    def test_getitem_list(self):
        edf = EcoDataFrame(self.gdf)
        edf = edf[["id"]]
        assert isinstance(edf, EcoDataFrame)
        assert len(edf) == 2

    def test_getitem_slice(self):
        edf = EcoDataFrame(self.gdf)
        edf = edf[slice(1)]
        assert isinstance(edf, EcoDataFrame)
        assert len(edf) == 1

    def test_astype(self):
        edf = EcoDataFrame({"a": [1], "geometry": [Point(0, 0)]})
        edf = edf.astype("object")
        assert isinstance(edf, EcoDataFrame)

    def test_merge(self):
        edf1 = EcoDataFrame({"lkey": ["foo", "bar"], "value": [1, 2]})
        edf2 = EcoDataFrame({"rkey": ["foo", "bar"], "value": [5, 6]})
        edf = edf1.merge(edf2, left_on="lkey", right_on="rkey")
        assert isinstance(edf, EcoDataFrame)

    def test_dissolve(self):
        edf = EcoDataFrame({"a": [1], "geometry": [Point(0, 0)]})
        edf = edf.dissolve(by="a")
        assert isinstance(edf, EcoDataFrame)

    @pytest.mark.filterwarnings("ignore::FutureWarning")
    def test_explode(self):
        edf = EcoDataFrame({"a": [1], "geometry": [Point(0, 0)]})
        edf = edf.explode()
        assert isinstance(edf, EcoDataFrame)

    def test_concat(self):
        edf1 = EcoDataFrame({"a": [1], "geometry": [Point(0, 0)]})
        edf2 = EcoDataFrame({"a": [2], "geometry": [Point(1, 1)]})
        edf = pd.concat([edf1, edf2])
        assert isinstance(edf, EcoDataFrame)

    def test_plot_series(self, mocker):
        series_plot_mock = mocker.patch.object(pd.Series, "plot")
        edf = EcoDataFrame({"x": [1]})
        edf["x"].plot()
        series_plot_mock.assert_called_once_with()

    def test_plot_geoseries(self, mocker):
        geoseries_plot_mock = mocker.patch.object(gpd.GeoSeries, "plot")
        edf = EcoDataFrame({"geometry": [Point(0, 0)]})
        edf["geometry"].plot()
        geoseries_plot_mock.assert_called_once_with()

    def test_plot_dataframe(self, mocker):
        dataframe_plot_mock = mocker.patch.object(pd.DataFrame, "plot")
        edf = EcoDataFrame({"x": [1], "y": [1]})
        edf[["x", "y"]].plot()
        dataframe_plot_mock.assert_called_once_with()

    def test_plot_geodataframe(self, mocker):
        geodataframe_plot_mock = mocker.patch.object(gpd.GeoDataFrame, "plot")
        edf = EcoDataFrame({"x": [1], "geometry": [Point(0, 0)]})
        subset = edf[["x", "geometry"]]
        subset.plot(column="x")
        geodataframe_plot_mock.assert_called_once_with(subset, column="x")



import numpy as np
import pandas as pd
import plotly.graph_objs as go
import shapely
from plotly.subplots import make_subplots
from sklearn.neighbors import KernelDensity


class EcoPlotData:
    def __init__(self, grouped, x_col="x", y_col="y", groupby_style=None, **style):
        self.grouped = grouped
        self.x_col = x_col
        self.y_col = y_col
        self.groupby_style = {} if groupby_style is None else groupby_style
        self.style = style

        # Plotting Defaults
        self.style["mode"] = self.style.get("mode", "lines+markers")


def ecoplot(
    data,
    title="",
    out_path=None,
    subplot_height=100,
    subplot_width=700,
    vertical_spacing=0.001,
    annotate_name_pos=(0.01, 0.99),
    y_title_2=None,
    layout_kwargs=None,
    **make_subplots_kwargs,
):
    groups = sorted(list(set.union(*[set(datum.grouped.groups.keys()) for datum in data])))
    datum_1 = data[0]
    datum_2 = None
    for datum in data[1:]:
        if datum.y_col != datum_1.y_col:
            datum_2 = datum
            break

    n = len(groups)
    fig_height = n * subplot_height + 2 * subplot_height

    fig = make_subplots(
        **{
            **dict(
                rows=n,
                cols=1,
                shared_xaxes="all",
                vertical_spacing=vertical_spacing,
                x_title=data[0].x_col,
                y_title=data[0].y_col,
                row_heights=list(np.repeat(subplot_height, n)),
                column_widths=[subplot_width],
                specs=[[{"secondary_y": datum_2 is not None}]] * len(groups),
            ),
            **make_subplots_kwargs,
        }
    )

    for i, name in enumerate(groups, 1):
        for datum in data:
            try:
                df = datum.grouped.get_group(name)
            except KeyError:
                continue

            timeseries = go.Scatter(
                x=df[datum.x_col],
                y=df[datum.y_col],
                name=name,
                **{**datum.style, **datum.groupby_style.get(name, {})},
            )
            fig.add_trace(
                timeseries,
                row=i,
                col=1,
                secondary_y=datum.y_col is not datum_1.y_col,
            )

    fig.update_xaxes(tickformat="%b-%Y")

    fig.layout.annotations[1]["font"]["color"] = datum_1.style.get("line", {}).get("color", "black")

    if datum_2 is not None:
        fig.layout.annotations = list(fig.layout.annotations) + [
            go.layout.Annotation(
                {
                    "font": {
                        "size": 16,
                        "color": datum_2.style.get("line", {}).get("color", "black"),
                    },
                    "showarrow": False,
                    "text": y_title_2 or datum_2.y_col,
                    "textangle": -90,
                    "x": 1,
                    "xanchor": "right",
                    "xref": "paper",
                    "xshift": +40,
                    "y": 0.5,
                    "yanchor": "middle",
                    "yref": "paper",
                },
            )
        ]

    fig["layout"].update(
        title={
            "text": title,
            "xanchor": "center",
            "x": 0.5,
        },
        height=fig_height,
    )

    fig.update_layout(**{**dict(showlegend=False, autosize=False), **(layout_kwargs or {})})

    if annotate_name_pos is not None:
        for i, name in enumerate(groups, 1):
            fig.add_annotation(
                text=name,
                showarrow=False,
                xref="x domain",
                yref="y domain",
                x=annotate_name_pos[0],
                y=annotate_name_pos[1],
                row=i,
                col=1,
            )

    if out_path is not None:
        fig.write_image(out_path, height=n * subplot_height)

    return fig


def add_seasons(fig, season_df):
    fig = make_subplots(figure=fig, specs=[[{"secondary_y": True}]])
    fig.add_trace(
        go.Scatter(
            x=pd.concat([season_df.start, season_df.end]).sort_values(),
            y=season_df.season.repeat(2),
            fill="tozeroy",
            fillcolor="rgba(0,0,255,0.1)",
            mode="none",
        ),
        secondary_y=True,
    )
    fig.update_yaxes(categoryorder="array", categoryarray=["dry", "wet"])
    return fig


def mcp(relocations):
    relocations = relocations.to_crs(relocations.estimate_utm_crs())

    areas = []
    times = []
    total = shapely.geometry.GeometryCollection()
    for time, obs in relocations.groupby(pd.Grouper(key="fixtime", freq="1D"), as_index=False):
        if obs.size:
            total = total.union(obs.geometry.unary_union).convex_hull
            areas.append(total.area)
            times.append(time)

    areas = np.array(areas)
    times = np.array(times)
    times[0] = relocations["fixtime"].iat[0]
    times[-1] = relocations["fixtime"].iat[-1]

    fig = go.FigureWidget()

    fig.add_trace(go.Scatter(x=times, y=areas / (1000**2)))

    fig.update_layout(
        margin_b=15,
        margin_l=50,
        margin_r=10,
        margin_t=25,
        title="MCP Asymptote",
        yaxis_title="MCP Area (km^2)",
        showlegend=False,
    )

    return fig


def nsd(relocations):
    relocations = relocations.to_crs(relocations.estimate_utm_crs())

    times = relocations["fixtime"]
    distances = relocations.distance(relocations.geometry[0]) ** 2

    fig = go.FigureWidget()

    fig.add_trace(go.Scatter(x=times, y=distances / (1000**2)))

    fig.update_layout(
        margin_b=15,
        margin_l=50,
        margin_r=10,
        margin_t=25,
        title="Net Square Displacement (NSD)",
        yaxis_title="NSD (km^2)",
        showlegend=False,
    )

    return fig


def speed(trajectory):
    times = np.column_stack(
        [
            trajectory["segment_start"],
            trajectory["segment_start"],
            trajectory["segment_end"],
            trajectory["segment_end"],
        ]
    ).flatten()
    speeds = np.column_stack(
        [
            np.zeros(len(trajectory)),
            trajectory["speed_kmhr"],
            trajectory["speed_kmhr"],
            np.zeros(len(trajectory)),
        ]
    ).flatten()

    fig = go.FigureWidget()

    fig.add_trace(go.Scatter(x=times, y=speeds))

    fig.update_layout(
        margin_b=15,
        margin_l=50,
        margin_r=10,
        margin_t=25,
        title="Speed",
        yaxis_title="Speed (km/h)",
        showlegend=False,
    )

    return fig


def plot_seasonal_dist(ndvi_vals, cuts, bandwidth=0.05, output_file=None):
    x = ndvi_vals.sort_values().to_numpy().reshape(-1, 1)
    kde = KernelDensity(kernel="gaussian", bandwidth=bandwidth).fit(x)
    dens = np.exp(kde.score_samples(x))
    fig = go.Figure(
        data=go.Scatter(
            x=x.ravel(),
            y=dens,
            fill=None,
            showlegend=False,
            mode="lines",
            line={
                "width": 1,
                "shape": "spline",
            },
        )
    )

    [
        fig.add_vline(
            x=i, line_width=3, line_dash="dash", line_color="red", annotation_text=" Cut Val: {:.2f}".format(i)
        )
        for i in cuts[1:-1]
    ]
    fig.update_layout(xaxis_title="NDVI")
    if output_file:
        fig.write_image(output_file)
    return fig



import typing

import geopandas as gpd
import mapclassify
import pandas as pd
import shapely

import ecoscope.base


class SpeedDataFrame(ecoscope.base.EcoDataFrame):
    @classmethod
    def from_trajectory(
        cls,
        trajectory: ecoscope.base.Trajectory,
        classification_method: str = "equal_interval",
        num_classes: int = 6,
        bins: typing.List = None,
        speed_colors: typing.List = None,
    ):
        if not bins:
            bins = apply_classification(trajectory.speed_kmhr, num_classes, cls_method=classification_method)

        if not speed_colors:
            speed_colors = default_speed_colors

        speed_colors = speed_colors[: len(bins) - 1]

        speed_df = cls(
            geometry=gpd.GeoSeries(
                trajectory.geometry.values,
                index=pd.Index(
                    pd.cut(x=trajectory.speed_kmhr, bins=bins, labels=speed_colors, include_lowest=True),
                    name="speed_colour",
                ),
            )
            .groupby(level=0)
            .apply(lambda gs: shapely.multilinestrings(gs)),
            crs=trajectory.crs,
        )
        speed_df.reset_index(drop=False, inplace=True)
        speed_df["label"] = _speedmap_labels(bins)
        speed_df.sort_values("speed_colour", inplace=True)

        return speed_df


def _speedmap_labels(bins):
    return [f"{bins[i]:.1f} - {bins[i + 1]:.1f} km/hr" for i in range(len(bins) - 1)]


def apply_classification(x, k, cls_method="natural_breaks", multiples=[-2, -1, 1, 2]):
    """
    Function to select which classifier to apply to the speed distributed data.

    Parameters
    ----------
    x : np.ndarray
        The input array to be classified. Must be 1-dimensional
    k : int
        Number of classes required.
    cls_method : str
        Classification method
    multiples : Listlike
        The multiples of the standard deviation to add/subtract from the sample mean to define the bins. defaults=
    """

    classifier = classification_methods.get(cls_method)
    if not classifier:
        return

    map_classifier = classifier(x, multiples) if cls_method == "std_mean" else classifier(x, k)
    edges, _, _ = mapclassify.classifiers._format_intervals(map_classifier, fmt="{:.2f}")
    return [float(i) for i in edges]


default_speed_colors = [
    "#1a9850",
    "#91cf60",
    "#d9ef8b",
    "#fee08b",
    "#fc8d59",
    "#d73027",
]

classification_methods = {
    "equal_interval": mapclassify.EqualInterval,
    "natural_breaks": mapclassify.NaturalBreaks,
    "quantile": mapclassify.Quantiles,
    "std_mean": mapclassify.StdMean,
    "max_breaks": mapclassify.MaximumBreaks,
    "fisher_jenks": mapclassify.FisherJenks,
}
